/**
  Tiny OpenGL/GLUT particle system example program
  
  This version:
  	- Uses second-order dynamics
	- Treats "Z" as up (the right choice!)
	- Allows nice camera rotation using three vectors
	- Allows camera motion with "WASD" keys
  
  Dr. Orion Sky Lawlor, olawlor@acm.org, 2009-01-21 (Public Domain)
*/
#include "GL/glew.h" /* GL extension wrangler: http://glew.sourceforge.net/ */
#include "GL/glut.h" /* windows, mice, etc.  http://freeglut.sourceforge.net */
#include <stdlib.h> /* for "rand" */
#include <stdio.h> /* for printf */
#include "osl/vec4.h" /* 3D vector class */
#include "ogl/glsl.h" /* programmable shaders */
#include "ogl/glsl.cpp"

#define VEC3_TO_XYZ(v) (v).x,(v).y,(v).z /* convert a vec3 to three float arguments */
#include <vector> /* std::vector container */

/* Random Float: Return a random number between 0 and 1 */
float rf(void) { return (rand()&0xfff)*(1.0/0xfff); }
/* Random Vector: return a random vector in 3D, all axes between 0 and 1 */
vec3 rv(void) {return vec3(rf(),rf(),rf());}

bool key_down[256]; /* if true, the corresponding key is down */

float rotate_x=0, rotate_y=0;

/* Return the wall clock time, in seconds */
double oglTimer(void) {return 0.001*glutGet(GLUT_ELAPSED_TIME);}

class particle {
public:
	vec3 P; /* position */
	vec3 V; /* velocity */
	vec3 C; /* color */
};


/* 3D orientation */
/* Orthonormal coordinate frame */
class ortho_frame {
public:
	/* Unit vectors pointing along axes of our frame.
		X cross Y is Z. (right-handed coordinate system)
	*/
	vec3 x,y,z;
	ortho_frame() :x(1,0,0), y(0,0,1), z(0,-1,0) {} /* point down Y axis */
	
	/* Reorient this coordinate frame by this far along in the X and Y axes.  
	   "dx" is the distance along the z axis to push the x axis; 
	   "dy" the distance along the z axis to push the y axis.
	*/
	void nudge(double dx,double dy) {
		x+=dx*z;
		y+=dy*z;
		orthonormalize();
	}
	
	/* Reconstruct an orthonormal frame from X and Y axes.
	  Y is primary, X is secondary, Z is tertiary.
	*/
	void orthonormalize(void) {
		y=normalize(y);
		z=normalize(cross(x,y));
		x=normalize(cross(y,z));
	}
};
ortho_frame orient; /* camera orientation */
vec3 camera(0,-1.0,0.5); /* (initial) location of camera */


/* Return the distance between these two pointers, in bytes,
itself expressed as a pointer.  Needed for glVertexPointer &c. */
void *ptr_diff(void *a,void *b) {
	return (void *)(long)((char *)a - (char *)b);
}

void display(void) /* Draw everything in our window */
{
	static double elapsed=0.0; /* last elapsed time (seconds) */
	glFinish(); double start_t=oglTimer();
	
	glClearColor(0.3,0.5,0.7,1.0); /* clear to a nice blue color */
	glClear(GL_COLOR_BUFFER_BIT|GL_DEPTH_BUFFER_BIT);
	
	glMatrixMode(GL_PROJECTION);
	glLoadIdentity();
	gluPerspective(70.0, /* degree FOV */
		glutGet(GLUT_WINDOW_WIDTH)/float(glutGet(GLUT_WINDOW_HEIGHT)),
		0.1,1000.0 /* Z range */
	);
	
	/* Physics and display timesteps: */
	static float dt=0.001, display_dt=0.0;
	static float sim_time=0.0;
	float vel=3.0*display_dt; /* meters camera motion per frame */
	if (key_down[(int)'w']) camera-=vel*orient.z;
	if (key_down[(int)'s']) camera+=vel*orient.z;
	if (key_down[(int)'a']) camera-=vel*orient.x;
	if (key_down[(int)'d']) camera+=vel*orient.x;
	
	/* Don't allow head tilting! */
	orient.x.z=0.0; 
	orient.y=cross(orient.z,orient.x);
	orient.orthonormalize();
	
	vec3 target=camera-orient.z; /* camera looks down *negative* Z */
	gluLookAt(VEC3_TO_XYZ(camera), 
		VEC3_TO_XYZ(target), 
		VEC3_TO_XYZ(orient.y)
	);
	
	glMatrixMode(GL_MODELVIEW);
	glLoadIdentity();
	
	// Draw in a ground plane
	float big=30; /* size of the ground */
	float floor=0.0; /* Z height of floor */
	glDepthMask(GL_FALSE); /* don't write to depth buffer (no Z fighting!) */
	
	glBegin(GL_QUADS);
	glColor4f(0.6,0.4,0.2,1.0); /* light, opaque brown */
	glVertex3f(-big,-big,floor);
	glVertex3f(+big,-big,floor);
	glVertex3f(+big,+big,floor);
	glVertex3f(-big,+big,floor);
	glEnd();

	// Draw in a 1-meter grid on the ground
	glEnable(GL_LINE_SMOOTH);
	glEnable(GL_BLEND);
	glBlendFunc(GL_SRC_ALPHA,GL_ONE_MINUS_SRC_ALPHA);
	glBegin(GL_LINES);
	glColor4f(0.0,0.0,0.0,0.1); /* transparent black -> grey */
	for (int v=-big;v<=big;v++) {
		glVertex3f(-big,v,floor); /* X lines */
		glVertex3f(+big,v,floor);
		glVertex3f(v,-big,floor); /* Y lines */
		glVertex3f(v,+big,floor);
	}
	glEnd();
	
	glDepthMask(GL_TRUE); /* depth writes back on */
	
/* Draw axes */
	for (int axis=0;axis<3;axis++) {
		vec3 color(0.0);
		color[axis]=1.0;
		glColor3fv(color);
		
		glBegin(GL_LINES);
		glVertex3f(0,0,0);
		vec3 end(0.0); end[axis]=1.0;
		glVertex3fv(end);
		glEnd();
		
		glRasterPos3fv(end);
		glutBitmapCharacter(GLUT_BITMAP_TIMES_ROMAN_24,'X'+axis);
	}

/* Particle state storage */
	static std::vector<particle> pts;
	if (pts.size()==0) { /* First time: initialize points */
		srand(1);
		for (int i=0;i<10000;i++) {
			particle p;
			p.P=rv(); 
			p.P.x=p.P.z; /* z == initial Y */
			p.V=0.3*(rv()-vec3(0.5)); 
			p.V.z=0.0; /* don't move along Z */
			p.C=rv();
			pts.push_back(p);
		}
		pts[90].P.z=1.0;
	}
	
	// This keeps track of one particle's Y position at various X times
	static std::vector<vec3> history;
	
	static float energy=0, m=1.0; /* mass and energy (for energy accounting) */
	int ps=pts.size();
	
	static bool first_time=true; /* for initial bc's! */
	
	/* Subcycle, so the physics runs at the right stepsize, 
	   yet still keeps up with the display. */
	double wall_time=oglTimer()*0.1; /* <- run physics at 1/10 realtime */
	while (sim_time<wall_time && !key_down[(int)' '])
	{ /* take one physics step */
		sim_time+=dt;
		energy=0;
		for (int i=0;i<ps;i++) {
			particle &p=pts[i];

		/* Apply physics */
			vec3 A(0.0,0.0,-1.0); /* gravity acceleration */

			float KE=0.5*m*dot(p.V,p.V);
			float PE=m*dot(A,-p.P);
			energy+=PE+KE;

			p.P+=dt*p.V+0.5*dt*dt*A; /* move position by velocity */
			p.V+=dt*A; /* accelerate particles downwards */

			if (p.P.z<floor) { /* now beneath the floor */
				p.V.z=fabs(p.V.z); /* move UP, fool! */
				if (first_time) /* enforce boundary conditions */
					p.P.z=floor; /* raise up to floor: gains energy */
			}
			float wall=0.0;
			if (p.P.x<wall) {
				p.V.x=+fabs(p.V.x);
			}
			wall=+1.0;
			if (p.P.x>wall) {
				p.V.x=-fabs(p.V.x);
			}
			if (i==90) history.push_back(vec3(sim_time,p.P.z,0.0));
		}

		first_time=false;
	}
	
/* Draw our points */
	glEnable(GL_DEPTH_TEST);
	glEnable(GL_POINT_SMOOTH);
	glPointSize(8.4);
	/* Make gl_PointSize work from GLSL */
        glEnable(GL_VERTEX_PROGRAM_POINT_SIZE);
	
	static GLhandleARB simplePoint=makeProgramObject(
	"varying vec4 drawColor;\n"
	"uniform float pixPerParticle;\n"
	"void main(void) {\n"
	"	drawColor = gl_Color;\n"
	"	gl_Position = gl_ModelViewProjectionMatrix * gl_Vertex;\n"
	"	gl_PointSize = pixPerParticle/gl_Position.w;\n"
	"}\n"
	,
	"varying vec4 drawColor;\n"
	"void main(void) {\n"
	"	gl_FragColor = drawColor; \n"
	"}\n"
	);
	glUseProgramObjectARB(simplePoint);
	glUniform1fARB(glGetUniformLocationARB(simplePoint,"pixPerParticle"),
		6.0e-3*glutGet(GLUT_WINDOW_HEIGHT));
	
	glBegin(GL_POINTS);
	for (int i=0;i<ps;i++) {
		particle &p=pts[i];
	/* Draw the particle */
		glColor3fv(p.C);
		glVertex3fv(p.P);
	}
	glEnd();
	

/* Shadows */
	/* Smooth shadow edges using GL Point Sprites 
	   (magically makes texcoords appear in fragment program) 
	*/
        glEnable(GL_POINT_SPRITE_ARB);
        glTexEnvi(GL_POINT_SPRITE_ARB,GL_COORD_REPLACE_ARB,GL_TRUE);

	static GLhandleARB shadow=makeProgramObject(
	"varying vec4 drawColor;\n"
	"void main(void) {\n"
	"	vec4 floorLoc=gl_Vertex;\n"
	"	floorLoc.z=0.0; /* slap point down on the ground */\n"
	"	gl_Position = gl_ModelViewProjectionMatrix * floorLoc;\n"
	"	float uppity=10.0*gl_Vertex.z;\n"
	"	if (uppity>1.0) gl_Position.x=1000.0; /* fling point away */\n"
	"	drawColor = clamp(vec4(0,0,0,0.5-uppity*uppity),0.0,1.0);\n"
	"	gl_PointSize = (5.0+10.0*uppity)/gl_Position.w;\n"
	"}\n"
	,
	"varying vec4 drawColor;\n"
	"void main(void) {\n"
	"	vec2 center=vec2(gl_TexCoord[0])-vec2(0.5);\n"
	"	gl_FragColor = clamp(\n"
	"		+1.0/(1.0+10.0*dot(center,center))\n"
	"		-1.0/(1.0+10.0*(0.25))\n"
	"	,0.0,1.0)*drawColor; \n"
	"}\n"
	);
	glUseProgramObjectARB(shadow);
	glDepthMask(GL_FALSE); /* no depth writes */
	//glColor4f(0,0,0,0.1);
	glBegin(GL_POINTS);
	for (int i=0;i<ps;i++) {
		const vec3 v=pts[i].P;
		glVertex3fv(v);
	}
	glEnd();
	glDepthMask(GL_TRUE);
        glDisable(GL_POINT_SPRITE_ARB);
	
	glUseProgramObjectARB(0);
	
// Plot the history buffer on top of everything else
	glLoadIdentity();
	glMatrixMode(GL_PROJECTION);
	glLoadIdentity();
	glDisable(GL_DEPTH_TEST);
	glColor3f(1,1,1);
	glBegin(GL_LINE_STRIP);
	for (unsigned int h=0;h<history.size();h++) {
		glVertex2f(history[h].x/sim_time*2.0-1.0,history[h].y);
	}
	glEnd();
	
	glutSwapBuffers(); /* display everything we've rendered so far */
	glutPostRedisplay(); /* continual animation */
	
	// Print frames per second
	glFinish(); elapsed=oglTimer()-start_t; /* stop the stopwatch */
	static double last_print=0.0; /* Periodically print out the time */
	if (start_t>last_print+0.1) {
		display_dt=elapsed;
		printf("energy = %g\n",energy);
		printf("%.1f frames/sec, %.2f ms/frame\n",
			1.0/elapsed,1000.0*elapsed);
		last_print=start_t;
	}
}

int mouse_x=0,mouse_y=0;
void mouse(int button,int state,int x,int y) 
{ /* mouse being pressed or released--save position for motion */
	mouse_x=x; mouse_y=y;
}
void motion(int x, int y) { 
	float scale=0.01; /* radians rotation/frame per mouse pixel motion */
	orient.nudge(
		-scale*(x-mouse_x), 
		scale*(y-mouse_y)
	);
	mouse_x=x; mouse_y=y; /* save old mouse positions */
}
void keyboard(unsigned char key,int x,int y) {
	key_down[key]=true;
}
void keyboardUp(unsigned char key,int x,int y) {
	key_down[key]=false;
}

int main(int argc,char **argv) {
	glutInit(&argc,argv);  /* set up GLUT */
	glutInitWindowPosition(0,60);
	glutInitWindowSize(1000,700);
	glutInitDisplayMode(GLUT_DOUBLE | GLUT_RGBA | GLUT_DEPTH);
	glutCreateWindow("Particle Demo (try WASD keys!)"); /* make a window */
	glewInit(); /* only AFTER create window! */
	glutDisplayFunc(display); /* here's how to draw the window */
	glutMouseFunc(mouse);
	glutMotionFunc(motion); 
	glutKeyboardFunc(keyboard);
	glutKeyboardUpFunc(keyboardUp);
	glutMainLoop(); /* start drawing */
}
